import click

from tq42.cli.utils.types import TQ42CliContext
from tq42.dataset import Dataset, list_all, DatasetSensitivityProto


@click.group("dataset")
def dataset_group() -> click.Group:
    """
    Class to manage datasets.

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Work_with_Datasets.html
    """
    pass


@dataset_group.command("create")
@click.option(
    "--proj",
    "proj_id",
    required=True,
    type=str,
    help="Attribute the command to a particular project",
)
@click.option(
    "--name", "name", required=True, type=str, help="The official name of the dataset"
)
@click.option(
    "--desc",
    "description",
    required=True,
    type=str,
    help="A friendly description of the dataset",
)
@click.option(
    "--sensitivity",
    "sensitivity",
    required=True,
    type=str,
    help="One of the following categories: PUBLIC, GENERAL, SENSITIVE, CONFIDENTIAL",
)
@click.option("--url", "url", required=False, type=str, help="A URL to the dataset")
@click.option(
    "--file",
    "file",
    required=False,
    type=str,
    help="A file (path) to upload to the dataset",
)
@click.pass_context
def create_dataset(
    ctx: TQ42CliContext,
    proj_id: str,
    name: str,
    description: str,
    sensitivity: str,
    url: str = None,
    file: str = None,
) -> None:
    """
    Create a dataset within a project.

    e.g.:
    - File upload: tq42 proj dataset create --proj 98ccb1d2-a3d0-48c8-b172-022f6db9be01 --name "Example Dataset Name" --desc "Example Description"  --file "/local/file/path/file.txt" --sensitivity "confidential"
    - Bucket transfer: tq42 proj dataset create --proj 98ccb1d2-a3d0-48c8-b172-022f6db9be01 --name "Example Dataset Name" --desc "Example Description"  --url "https://mydata.com/drive/my-drive" --sensitivity "confidential"

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Working_with_Datasets.html
    """

    click.echo(
        Dataset.create(
            client=ctx.obj.client,
            project_id=proj_id,
            name=name,
            description=description,
            url=url,
            file=file,
            sensitivity=DatasetSensitivityProto.Value(sensitivity.upper()),
        ).data
    )


@dataset_group.command("list")
@click.option(
    "--proj",
    "proj_id",
    required=True,
    type=str,
    help="Attribute the command to a particular project",
)
@click.pass_context
def list_datasets(ctx: TQ42CliContext, proj_id: str) -> None:
    """
    List all datasets within a project.

    e.g.: tq42 proj dataset list --proj 98ccb1d2-a3d0-48c8-b172-022f6db9be01

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Working_with_Datasets.html
    """
    click.echo(
        [
            dataset.data
            for dataset in list_all(client=ctx.obj.client, project_id=proj_id)
        ]
    )


@dataset_group.command("get")
@click.argument("dataset_id", required=True, type=str)
@click.pass_context
def get(ctx: TQ42CliContext, dataset_id: str) -> None:
    """
    Get a specific dataset.

    e.g.: tq42 proj dataset get 98ccb1d2-a3d0-48c8-b172-022f6db9be01

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Working_with_Datasets.html
    """
    click.echo(Dataset(client=ctx.obj.client, id=dataset_id).data)


@dataset_group.command(
    "export",
)
@click.argument("dataset_id", required=True, type=str)
@click.argument("directory_path", required=True, type=str)
@click.pass_context
def export(ctx: TQ42CliContext, dataset_id: str, directory_path: str) -> None:
    """
    Export a specific dataset.

    e.g.: tq42 proj dataset export 98ccb1d2-a3d0-48c8-b172-022f6db9be01 /Users/user1/Downloads

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Working_with_Datasets.html
    """
    dataset = Dataset(client=ctx.obj.client, id=dataset_id)
    dataset.export(directory_path=directory_path)



from __future__ import annotations

from google.protobuf import empty_pb2

from com.terraquantum.plan.v1.plan.check_functionality_request_pb2 import (
    CheckFunctionalityRequest,
    FunctionalityProto,
)

from typing import TYPE_CHECKING

# only import the stuff for type hints -> avoid circular imports
if TYPE_CHECKING:
    from tq42.client import TQ42Client


class Functionality:
    """
    Reference a functionality.
    """

    @staticmethod
    def check(
        client: TQ42Client, organization_id: str, functionality_type: str, version: str
    ) -> None:
        """
        Checks if a functionality with given version should be accessible for the given organization.
        """
        req = CheckFunctionalityRequest(
            organization_id=organization_id,
            functionality=FunctionalityProto(type=functionality_type, version=version),
        )
        res: empty_pb2.Empty = client.plan_client.CheckFunctionality(
            request=req, metadata=client.metadata
        )
        return res



class PrettyList(list):
    def __repr__(self):
        if not self:
            return "[]"

        type_name = type(self[0]).__name__
        items = "\n  ".join([item.__repr__() for item in self])
        return f"{type_name}List: [\n  {items}\n]"

    def __str__(self):
        return self.__repr__()



import sys
from typing import Literal

import click
import click.shell_completion


@click.command("completion")
@click.argument("shell", required=True, type=click.Choice(["bash", "zsh", "fish"]))
def generate_completion(shell: Literal["bash", "zsh", "fish"]) -> None:
    """
    Generate completion scripts for your shell.
    Supported shells are:
    - bash
    - zsh
    - fish

    To load completions:
        Bash:
            source <(tq42 completion bash)
        Zsh:
            source <(tq42 completion zsh)
        Fish:
            tq42 completion fish | source
    """

    context = click.get_current_context()

    shells = {
        "bash": click.shell_completion.BashComplete,
        "zsh": click.shell_completion.ZshComplete,
        "fish": click.shell_completion.FishComplete,
    }

    shell_class = shells[shell]
    if not shell_class:
        raise ValueError(f"Shell {shell} is not supported.")

    complete_var = f"_{context.parent.info_name.upper()}_COMPLETE"
    sys.stdout.write(
        shell_class(
            cli=context.command,
            prog_name=context.parent.info_name,
            ctx_args={},
            complete_var=complete_var,
        ).source()
    )



import os
import unittest

from tq42.utils import dirs


class TestDirs(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_code(self):
        self.assertTrue(os.path.exists(dirs.code()))
        self.assertTrue(os.path.isfile(dirs.code("__init__.py")))

    def test_testdata(self):
        self.assertTrue(os.path.exists(dirs.testdata()))
        self.assertTrue(os.path.isfile(dirs.testdata("README.md")))

    def test_cache(self):
        self.assertTrue(os.path.exists(dirs.cache_dir()))
        self.assertIsNotNone(dirs.cache_file())

    def test_text_files(self):
        self.assertTrue(os.path.exists(dirs.text_files_dir()))
        self.assertTrue(
            os.path.isfile(
                dirs.full_path(dirs.text_files_dir(), "unauthenticated_error.txt")
            )
        )
        self.assertTrue(
            os.path.isfile(
                dirs.full_path(
                    dirs.text_files_dir(), "insufficient_permission_error.txt"
                )
            )
        )

    def test_create_or_get_package_folder_path(self):
        config_dir = dirs.create_or_get_config_dir()
        self.assertTrue(os.path.exists(config_dir))
        self.assertEqual(dirs.CONFIG_FOLDER_NAME, os.path.basename(config_dir))

    def test_config(self):
        expected = os.path.join(dirs.create_or_get_config_dir(), "token.json")
        actual = dirs.config("token.json")
        self.assertEqual(expected, actual)



from __future__ import annotations

import json
import time
from typing import Optional, List, Mapping, Any, Union

from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    ExperimentRunStatusProto,
)

# important for re-export
from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    HardwareProto,
)
from com.terraquantum.experiment.v3alpha2.experimentrun.cancel_experiment_run_request_pb2 import (
    CancelExperimentRunRequest,
)
from com.terraquantum.experiment.v3alpha2.experimentrun.create_experiment_run_request_pb2 import (
    CreateExperimentRunRequest,
)
from com.terraquantum.experiment.v3alpha2.experimentrun.experiment_run_pb2 import (
    ExperimentRunProto,
)
from com.terraquantum.experiment.v3alpha2.experimentrun.get_experiment_run_request_pb2 import (
    GetExperimentRunRequest,
)
from com.terraquantum.experiment.v3alpha2.experimentrun.list_experiment_runs_pb2 import (
    ListExperimentRunsRequest,
)
from com.terraquantum.experiment.v3alpha2.experimentrun.list_experiment_runs_pb2 import (
    ListExperimentRunsResponse,
)
from google.protobuf import struct_pb2
from google.protobuf.json_format import MessageToJson, ParseDict, MessageToDict

from tq42.client import TQ42Client
from tq42.utils.exception_handling import handle_generic_sdk_errors
from tq42.exceptions import ExperimentRunCancelError, ExceedRetriesError
from tq42.utils.pretty_list import PrettyList


class ExperimentRun:
    """
    Reference an existing experiment run.

    :param client: a client instance
    :param id: the id of the existing experiment run
    :param data: only used internally
    """

    _client: TQ42Client
    id: str
    """ID of the experiment run"""
    data: ExperimentRunProto
    """Object containing all attributes of the experiment run"""

    def __init__(
        self, client: TQ42Client, id: str, data: Optional[ExperimentRunProto] = None
    ):
        self._client = client
        self.id = id

        if data:
            self.data = data
        else:
            self.data = self._get_data()

    def __repr__(self) -> str:
        return f"<ExperimentRun Id={self.id}>"

    def __str__(self) -> str:
        return f"ExperimentRun: {MessageToJson(self.data, preserving_proto_field_name=True)}"

    @property
    def completed(self) -> bool:
        """
        Check if the experiment run is completed

        :returns: true if the experiment run is in the state `COMPLETED`
        """

        return self.data.status == ExperimentRunStatusProto.COMPLETED

    @property
    def result(self) -> Optional[dict[str, Any]]:
        """
        Get the result of the experiment run if the run is completed.

        If the result contains a results_string or if the result is a string, it will be parsed and returned.

        :returns: a dict with the result of the experiment run. If the run is not completed yet, returns `None`.
        """

        if not self.completed:
            return None

        result: Union[dict[str, Any], str] = MessageToDict(
            self.data.result.outcome
        ).get("result", {})
        if isinstance(result, str):
            return json.loads(result)
        elif "results_string" in result:
            return json.loads(result.get("results_string"))

        return result

    @property
    def outputs(self) -> Optional[dict[str, Any]]:
        """
        Get the outputs of the experiment run if the run is completed.

        :returns: a dict with the outputs of the experiment run. If the run is not completed yet, returns `None`.
        """

        if not self.completed:
            return None

        return MessageToDict(self.data.result.outcome).get("outputs", {})

    @handle_generic_sdk_errors
    def _get_data(self) -> ExperimentRunProto:
        """
        Gets a specific experiment run by id
        """
        get_exp_run_request = GetExperimentRunRequest(experiment_run_id=self.id)

        res = self._client.experiment_run_client.GetExperimentRun(
            request=get_exp_run_request, metadata=self._client.metadata
        )

        return res

    @staticmethod
    def from_proto(client: TQ42Client, msg: ExperimentRunProto) -> ExperimentRun:
        """
        Creates ExperimentRun instance from a protobuf message.

        :meta private:
        """
        return ExperimentRun(client=client, id=msg.id, data=msg)

    @staticmethod
    @handle_generic_sdk_errors
    def create(
        client: TQ42Client,
        algorithm: str,
        version: str,
        experiment_id: str,
        compute: HardwareProto,
        parameters: Mapping[str, Any],
    ) -> ExperimentRun:
        """
        Start a new experiment run in an experiment

        :param client: a client instance
        :param algorithm: name of the algorithm (e.g. `'TOY'`)
        :param version: version of the algorithm in the format `x.y.z`
        :param experiment_id: id of the experiment in which the run should be started
        :param compute: the hardware specification on which the run should be started (e.g. `HardwareProto.SMALL`)
        :param parameters: dict with parameters for the algorithm
        :returns: the created experiment run

        """

        request = CreateExperimentRunRequest(
            experiment_id=experiment_id,
            algorithm=algorithm,
            version=version,
            hardware=compute,
            metadata=ParseDict(parameters, struct_pb2.Struct()),
        )

        res: ExperimentRunProto = client.experiment_run_client.CreateExperimentRun(
            request=request, metadata=client.metadata
        )

        return ExperimentRun.from_proto(client=client, msg=res)

    @handle_generic_sdk_errors
    def check(self) -> ExperimentRun:
        """
        Update the state of the experiment run

        :returns: the updated experiment run
        """
        self.data = self._get_data()
        return self

    @handle_generic_sdk_errors
    def poll(
        self, tries=1000, initial_delay=1.0, delay=1.0, backoff=1.0
    ) -> ExperimentRun:
        """
        Monitor an experiment run until it completes, then automatically display the results (if there are no errors).

        :param tries: how many retries until the poll loop is cancelled (default: 1000)
        :param initial_delay: initial delay before starting poll loop (default: 1 second)
        :param delay: initial delay between retries (default: 1 second)
        :param backoff: backoff factor between retries (default: 1)
        :returns: the finished experiment run
        :raises: ExceedRetriesError if `tries` are exceeded
        """
        time.sleep(initial_delay)

        for _ in range(tries):
            self.data = self._get_data()
            if self.data.status in [
                ExperimentRunStatusProto.COMPLETED,
                ExperimentRunStatusProto.CANCELLED,
                ExperimentRunStatusProto.FAILED,
            ]:
                return self

            time.sleep(delay)
            delay *= backoff

        raise ExceedRetriesError(tries=tries)

    @handle_generic_sdk_errors
    def cancel(self) -> ExperimentRun:
        """
        Cancel a run that is QUEUED, PENDING, or RUNNING.

        :returns: the cancelled experiment run
        :raises: ExperimentRunCancelError if the experiment run is not queued, pending or running
        """
        try:
            cancel_exp_runs_response = CancelExperimentRunRequest(
                experiment_run_id=self.id
            )
            self._client.experiment_run_client.CancelExperimentRun(
                request=cancel_exp_runs_response, metadata=self._client.metadata
            )
            return self
        except Exception:
            raise ExperimentRunCancelError()


@handle_generic_sdk_errors
def list_all(client: TQ42Client, experiment_id: str) -> List[ExperimentRun]:
    """
    List all the runs within an experiment you have permission to view.

    :param client: a client instance
    :param experiment_id: id of the experiment
    :returns: a list of experiment runs
    """
    list_exp_run_request = ListExperimentRunsRequest(experiment_id=experiment_id)

    res: ListExperimentRunsResponse = client.experiment_run_client.ListExperimentRuns(
        request=list_exp_run_request, metadata=client.metadata
    )
    # TODO: It seems like currently the API returns `experiment_runs` instead of `experimentRuns` as in the protobufs
    return PrettyList(
        [
            ExperimentRun.from_proto(client=client, msg=experiment_run)
            for experiment_run in res.experiment_runs
        ]
    )



# ruff: noqa: F401
from tq42.exceptions.authentication_error import AuthenticationError
from tq42.exceptions.experiment_run_cancel_error import ExperimentRunCancelError
from tq42.exceptions.invalid_argument_error import InvalidArgumentError
from tq42.exceptions.no_default_error import NoDefaultError
from tq42.exceptions.permission_denied_error import PermissionDeniedError
from tq42.exceptions.retries_exceeded_error import ExceedRetriesError
from tq42.exceptions.tq42_api_error import TQ42APIError
from tq42.exceptions.unauthenticated_error import UnauthenticatedError



import unittest
from unittest import mock

from tq42.utils.environment import ConfigEnvironment


class TestConfigEnvironment(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def code_data_test(self, env: ConfigEnvironment):
        expected = {
            "client_id": env.client_id,
            "scope": env.scope,
            "audience": env.audience,
        }
        actual = env.code_data
        self.assertEqual(expected, actual)

        expected = {"Content-Type": "application/x-www-form-urlencoded"}
        actual = env.headers
        self.assertEqual(expected, actual)

    def token_data_test(self, env: ConfigEnvironment):
        expected = {
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": "DEVICE CODE",
            "client_id": env.client_id,
        }
        actual = env.token_data("DEVICE CODE")
        self.assertEqual(expected, actual)

    def test_int_urls(self):
        env = ConfigEnvironment(
            "int.terraquantum.io",
            "CLIENT ID",
            "openid profile email",
        )
        self.assertEqual("api.int.terraquantum.io", env.api_host)
        self.assertEqual(
            "https://graphql-gateway.int.terraquantum.io/graphql", env.audience
        )
        self.assertEqual(
            "https://auth.int.terraquantum.io/oauth/device/code", env.auth_url_code
        )
        self.assertEqual(
            "https://auth.int.terraquantum.io/oauth/token", env.auth_url_token
        )
        self.code_data_test(env)
        self.token_data_test(env)

    def test_staging_urls(self):
        env = ConfigEnvironment(
            "staging.terraquantum.io",
            "CLIENT ID",
            "openid profile email",
        )

        self.assertEqual("api.staging.terraquantum.io", env.api_host)
        self.assertEqual(
            "https://graphql-gateway.staging.terraquantum.io/graphql", env.audience
        )
        self.assertEqual(
            "https://auth.staging.terraquantum.io/oauth/device/code", env.auth_url_code
        )
        self.assertEqual(
            "https://auth.staging.terraquantum.io/oauth/token", env.auth_url_token
        )
        self.code_data_test(env)
        self.token_data_test(env)

    @mock.patch.dict(
        "os.environ",
        {
            "TQ42_BASE_URL": "base_url",
            "TQ42_CLIENT_ID": "client_id",
            "TQ42_SCOPE": "scope",
        },
    )
    def test_environment_name(self):
        env = ConfigEnvironment.from_env()

        self.assertEqual("base_url", env.base_url)
        self.assertEqual("client_id", env.client_id)
        self.assertEqual("scope", env.scope)

    def test_default_environment(self):
        env = ConfigEnvironment.from_env()

        self.assertEqual("terraquantum.io", env.base_url)
        self.assertEqual("gvBa4BHKOTlotDuE6E2HSQBzBDlM00F4", env.client_id)
        self.assertEqual("openid profile email offline_access tq42", env.scope)



import OptimizationTestFunctions as otf
import numpy as np
from pytest import mark, fixture

from tq42.channel import Channel, Ask, Tell
from tq42.experiment_run import ExperimentRun, HardwareProto
from tq42.functional_tests.functional_test_config import FunctionalTestConfig
from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    ExperimentRunStatusProto,
)


@fixture
def functional_test_config():
    return FunctionalTestConfig()


@mark.asyncio
async def test_channel_create(functional_test_config):
    channel = await Channel.create(client=functional_test_config.get_client())
    assert channel.id is not None


@mark.asyncio
async def test_exp_run_with_channel(functional_test_config):
    channel = await Channel.create(client=functional_test_config.get_client())
    cva_params = {
        "objectives": [{"name": "Ackley", "aim_type": "MINIMIZE"}],
        "variables": [
            {"name": "x1", "info_real": {"lower_bound": -1.0, "upper_bound": 1.0}},
            {"name": "x2", "info_real": {"lower_bound": -1.0, "upper_bound": 1.0}},
        ],
        "func_eval_worker_channel_id": channel.id,
        "parameters": {"max_generation": 3, "mue": 15, "lambda": 100},
    }

    exp_run = ExperimentRun.create(
        client=functional_test_config.get_client(),
        algorithm="CVA_OPT",
        version="0.1.0",
        experiment_id=functional_test_config.exp,
        compute=HardwareProto.SMALL,
        parameters={"parameters": cva_params, "inputs": {}},
    )

    async def callback(ask: Ask) -> Tell:
        dim = len(ask.headers)
        func = otf.Ackley(dim)
        y = []
        for parameter in ask.parameters:
            y.append(float(func(np.array(parameter.values))))

        return Tell(parameters=ask.parameters, headers=ask.headers, results=y)

    def success():
        poll_result = exp_run.poll()
        assert ExperimentRunStatusProto.COMPLETED == poll_result.data.status

    await channel.connect(
        callback=callback,
        finish_callback=success,
        max_duration_in_sec=None,
        message_timeout_in_sec=300,
    )

    assert exp_run.data.status == ExperimentRunStatusProto.COMPLETED



import unittest

from tq42.functional_tests.functional_test_config import FunctionalTestConfig
from tq42.utils.cache import clear_cache, get_current_value

from tq42.utils.environment import environment_default_set


class TestFunctionalCacheUtils(unittest.TestCase, FunctionalTestConfig):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_default_ids(self):
        environment_default_set(client=self.get_client())
        org_id = get_current_value("org")
        proj_id = get_current_value("proj")
        self.assertIsNotNone(org_id, "org_id is None")
        self.assertIsNotNone(proj_id, "proj_id is None")
        clear_cache()



from tq42.exceptions.tq42_api_error import TQ42APIError

_MESSAGE = """
Unable to Execute: `{}`

We were unable to automatically detect the default org and/or proj ID required to run this command.
Please set an org and proj yourself by running the following commands:

`tq42 org set <ORG_ID>`
`tq42 proj set <PROJ_ID>`

These IDs can be either found on terraquantum.io or via the corresponding CLI commands `tq42 org list` / `tq42 proj list`.
For a list of available commands, type `tq42 --help` or consult the documentation at tq42.com/help.
""".strip()


class NoDefaultError(TQ42APIError):
    """
    Raised when the default organisation and/or project cannot be detected

    Attributes:
        command (str): the command for which the error occurred
    """

    def __init__(self, command: str):
        self.command = command

    def __str__(self):
        return _MESSAGE.format(self.command)



import unittest

from tq42.utils import dirs, file_handling
from tq42.utils.cache import (
    clear_cache,
    write_key_value_to_cache,
    get_current_value,
)
from tq42.functional_tests.functional_test_config import FunctionalTestConfig


class TestCacheUtils(unittest.TestCase, FunctionalTestConfig):
    def setUp(self):
        pass

    def test_get_current_id(self):
        clear_cache()
        write_key_value_to_cache("org", "RANDOM_ORG_ID")
        org_id = get_current_value("org")
        self.assertEqual("RANDOM_ORG_ID", org_id)
        clear_cache()

    def test_clear_cache(self):
        write_key_value_to_cache("org", "RANDOM_ORG_ID")
        content = file_handling.read_file(dirs.cache_file())
        self.assertNotEqual(content, "")
        clear_cache()
        content = file_handling.read_file(dirs.cache_file())
        self.assertEqual(content, "")



import unittest
from pytest import mark

from tq42.experiment_run import ExperimentRun, HardwareProto
from tq42.functional_tests.functional_test_config import FunctionalTestConfig

from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    ExperimentRunStatusProto,
)


@mark.poll
class TestFunctionalPollingTQ42API(unittest.TestCase, FunctionalTestConfig):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_exp_run_poll(self):
        parameters = {
            "parameters": {
                "n": 1,
                "r": 1,
                "msg": "correct",
            },
            "inputs": {},
        }

        exp_run = ExperimentRun.create(
            client=self.get_client(),
            algorithm="TOY",
            version="0.1.0",
            experiment_id=self.exp,
            compute=HardwareProto.SMALL,
            parameters=parameters,
        )
        poll_result = exp_run.poll()
        self.assertEqual(ExperimentRunStatusProto.COMPLETED, poll_result.data.status)
        self.assertEqual(ExperimentRunStatusProto.COMPLETED, exp_run.data.status)



import unittest

from tq42.project import Project
from tq42.utils.cache import clear_cache
from tq42.functional_tests.functional_test_config import FunctionalCLITestConfig


class TestFunctionalInvokeCliExperimentGroup(
    unittest.TestCase, FunctionalCLITestConfig
):
    def test_exp_list(self):
        result = self.runner.invoke(
            self.cli_entry, ["exp", "list", "--proj", self.proj]
        )

        assert result.exit_code == 0
        assert result.output is not None

        # need a default set for just running exp list
        Project(client=self.get_client(), id=self.proj).set()
        result = self.runner.invoke(self.cli_entry, ["exp", "list"])
        assert result.exit_code == 0
        assert result.output is not None
        # clear the defaults to have each test being individual
        clear_cache()

    def test_exp_set_friendly_name(self):
        result = self.runner.invoke(
            self.cli_entry,
            ["exp", "set-friendly-name", "FRIENDLY_NAME", "--exp=" + self.exp],
        )
        assert result.exit_code == 0
        assert result.output is not None



from tq42.exceptions.tq42_api_error import TQ42APIError

_MESSAGE = """
Unauthorized Access: Insufficient Permissions
You do not have the necessary privileges to execute the requested command. Double check that the resource exists and contact your system administrator or request the appropriate access level.
Suggestions:
Verify that you have the required permissions to perform the command.
If you believe you should have access, contact your system administrator or request the necessary privileges.
Ensure that you are logged in with the correct user account.
""".strip()


class PermissionDeniedError(TQ42APIError):
    """
    Raised when the user does not have access to the requested resource
    """

    def __str__(self):
        return _MESSAGE



import keyring
from keyring.errors import NoKeyringError, InitError, PasswordSetError, KeyringLocked

from tq42.utils import file_handling


def save_token(service_name: str, backup_save_path: str, token: str) -> str:
    try:
        save_location = "keyring"
        keyring.set_password(
            service_name=service_name,
            username="username",
            password=token,
        )
        return save_location

    except (NoKeyringError, InitError, PasswordSetError):
        file_handling.write_to_file(backup_save_path, token)
        return backup_save_path


def get_token(service_name: str, backup_save_path: str) -> str:
    try:
        return keyring.get_password(
            service_name=service_name,
            username="username",
        )

    except (NoKeyringError, InitError, KeyringLocked):
        return file_handling.read_file(backup_save_path)



import asyncio
from typing import Optional


class AsyncTimedIterable:
    def __init__(self, iterable, timeout: Optional[int] = None):
        class AsyncTimedIterator:
            def __init__(self):
                self._iterator = iterable.__aiter__()

            async def __anext__(self):
                try:
                    result = await asyncio.wait_for(
                        self._iterator.__anext__(), timeout=timeout
                    )
                    return result
                except asyncio.TimeoutError as e:
                    raise e

        self._factory = AsyncTimedIterator

    def __aiter__(self):
        return self._factory()



from tq42.exceptions.tq42_api_error import TQ42APIError

_MESSAGE = """
Unable to Execute: `{}`

We were unable to execute the given command as it violates certain constraints for the given resource:
{}
""".strip()


class InvalidArgumentError(TQ42APIError):
    """
    Raised when providing wrong arguments / parameters.

    Attributes:
        command (str): Where the error occurred.
        details (str): Details about the error.
    """

    def __init__(self, command: str, details: str):
        self.command = command
        self.details = details

    def __str__(self):
        return _MESSAGE.format(self.command, self.details)



import click

import tq42.cli.utils.cli_functions as cli
from .model_group import model_group
from .dataset_group import dataset_group
from .utils.types import TQ42CliContext


@click.group("proj")
def project_group() -> click.Group:
    """
    Class to manage projects

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Setting_Up_Your_Environment.html#check-your-current-organization-and-project-settings
    """
    pass


@project_group.command("list")
@click.option(
    "--org",
    "org_id",
    required=False,
    type=str,
    help="Attribute the command to a particular organization",
)
@click.pass_context
def list_by_org(ctx: TQ42CliContext, org_id: str) -> None:
    """
    List all the projects you have permission to view within the organization that is currently set.

    e.g.: tq42 proj list --org b0edfd26-0817-4818-a278-17ef6c14e3a5

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Setting_Up_Your_Environment.html#list-all-projects
    """
    click.echo(cli.list_proj_by_org(ctx.obj.client, org_id))


@project_group.command("show")
@click.pass_context
def show_default(ctx: TQ42CliContext) -> None:
    """
    Show the current organization, project ID and associated project friendly name (if any).

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Setting_Up_Your_Environment.html#check-your-current-organization-and-project-settings
    """
    click.echo(cli.proj_show(ctx.obj.client))


@project_group.command("set")
@click.argument("proj_id", required=True, type=str)
@click.pass_context
def set_default(ctx: TQ42CliContext, proj_id: str) -> None:
    """
    Change the active project.

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Setting_Up_Your_Environment.html#changing-your-workspace-to-a-different-organization-or-project
    """
    click.echo(cli.set_project(ctx.obj.client, proj_id=proj_id))


@project_group.command("set-friendly-name")
@click.argument("friendly_name", required=True, type=str)
@click.option(
    "--proj",
    "proj_id",
    required=True,
    type=str,
    help="Attribute the command to a particular project",
)
@click.pass_context
def proj_set_friendly_name(
    ctx: TQ42CliContext, friendly_name: str, proj_id: str
) -> None:
    """
    Set a friendly name for a project.

    tq42 proj set-friendly-name NEW_FRIENDLY_NAME --proj b0edfd26-0817-4818-a278-17ef6c14e3a5

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Setting_Up_Your_Environment.html#setting-friendly-names-for-projects-and-experiments
    """
    click.echo(
        cli.proj_update(ctx.obj.client, proj_id=proj_id, friendly_name=friendly_name)
    )


project_group.add_command(model_group)
project_group.add_command(dataset_group)



import click

import tq42.cli.utils.cli_functions as cli
import tq42.utils.cache as cache_utils
from .experiment_run_group import exp_run_group
from .utils.types import TQ42CliContext


@click.group("exp")
def experiment_group() -> click.Group:
    """
    Class to manage experiments

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Submitting_and_Monitoring_a_Run.html#
    """
    pass


@experiment_group.command("list")
@click.option(
    "--proj",
    "proj_id",
    required=False,
    type=str,
    help="Attribute the command to a particular project",
)
@click.pass_context
def list_by_proj(ctx: TQ42CliContext, proj_id: str) -> None:
    """
    List all the experiments you have permission to view within a specific project.

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Setting_Up_Your_Environment.html#list-all-experiments
    """

    # TODO: Need to display error message to stderr when no proj is specified or set.
    #  Should do this by raising an exception and displaying from main function

    error_msg = "No project id is set. \nExample: tq42 exp list --proj b0edfd26-0817-4818-a278-17ef6c14e3a5"

    if proj_id is None:
        proj = cache_utils.get_current_value("proj")
        if proj == "":
            click.echo(error_msg)
            return

    click.echo(cli.list_exp_by_proj(ctx.obj.client, proj_id))


@experiment_group.command("set-friendly-name")
@click.argument("friendly_name", required=True, type=str)
@click.option(
    "--exp",
    "exp_id",
    required=True,
    type=str,
    help="Attribute the command to a particular experiment",
)
@click.pass_context
def set_friendly_name(ctx: TQ42CliContext, friendly_name: str, exp_id: str) -> None:
    """
    Set a friendly name for an experiment.

    e.g.: tq42 exp set-friendly-name NEW_FRIENDLY_NAME --exp b0edfd26-0817-4818-a278-17ef6c14e3a5

    https://docs.tq42.com/en/latest/CLI_Developer_Guide/Setting_Up_Your_Environment.html#setting-friendly-names-for-projects-and-experiments
    """
    click.echo(cli.exp_update(ctx.obj.client, exp_id=exp_id, name=friendly_name))


experiment_group.add_command(exp_run_group)



import random
import unittest


class TestExample(unittest.TestCase):
    def setUp(self):
        self.seq = list(range(10))

    def tearDown(self):
        pass

    def test_shuffle(self):
        # make sure the shuffled sequence does not lose any elements
        random.shuffle(self.seq)
        self.seq.sort()
        self.assertEqual(self.seq, list(range(10)))

        # should raise an exception for an immutable sequence
        self.assertRaises(TypeError, random.shuffle, (1, 2, 3))

    def test_choice(self):
        element = random.choice(self.seq)
        self.assertTrue(element in self.seq)

    def test_sample(self):
        with self.assertRaises(ValueError):
            random.sample(self.seq, 20)
        for element in random.sample(self.seq, 5):
            self.assertTrue(element in self.seq)



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Optional

from tq42.organization import Organization, list_all as list_all_organizations
from tq42.project import Project, list_all as list_all_projects
from tq42.utils import dirs, file_handling
from tq42.utils.cache import clear_cache

from typing import TYPE_CHECKING

# only import the stuff for type hints -> avoid circular imports
if TYPE_CHECKING:
    from tq42.client import TQ42Client

_DEFAULT_BASE_URL = "terraquantum.io"
_DEFAULT_CLIENT_ID = "gvBa4BHKOTlotDuE6E2HSQBzBDlM00F4"
_DEFAULT_SCOPE = "openid profile email offline_access tq42"


@dataclass
class ConfigEnvironment:
    """
    Configuration environment for the TQ42 SDK
    """

    base_url: str
    client_id: str
    scope: str

    @property
    def api_host(self):
        return "api.{}".format(self.base_url)

    @property
    def channels_host(self):
        return f"channels.{self.base_url}"

    @property
    def auth_url_token(self):
        return f"https://auth.{self.base_url}/oauth/token"

    @property
    def auth_url_code(self):
        return f"https://auth.{self.base_url}/oauth/device/code"

    @property
    def audience(self):
        return f"https://graphql-gateway.{self.base_url}/graphql"

    @property
    def client_credential_flow_audience(self):
        return f"https://api.{self.base_url}"

    @property
    def headers(self):
        return {"Content-Type": "application/x-www-form-urlencoded"}

    @property
    def code_data(self):
        return {
            "client_id": self.client_id,
            "scope": self.scope,
            "audience": self.audience,
        }

    def token_data(self, device_code: str):
        return {
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": device_code,
            "client_id": self.client_id,
        }

    def refresh_token_data(self, refresh_token: str):
        return {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": self.client_id,
        }

    def client_credentials_data(
        self, client_id: str, client_secret: str, audience: str
    ):
        return {
            "client_id": client_id,
            "client_secret": client_secret,
            "grant_type": "client_credentials",
            "audience": audience,
        }

    @staticmethod
    def from_env() -> "ConfigEnvironment":
        base_url = os.getenv("TQ42_BASE_URL", _DEFAULT_BASE_URL)
        client_id = os.getenv("TQ42_CLIENT_ID", _DEFAULT_CLIENT_ID)
        scope = os.getenv("TQ42_SCOPE", _DEFAULT_SCOPE)

        return ConfigEnvironment(base_url=base_url, client_id=client_id, scope=scope)


def get_environment() -> str:
    content = file_handling.read_file(dirs.cache_file())
    return content


def environment_clear() -> str:
    clear_cache()
    return get_environment()


def environment_default_set(
    client: TQ42Client, organization: Optional[Organization] = None
) -> str:
    if organization is None:
        organization = Organization.get_default_org(client=client)

    organization.set()

    content = file_handling.read_file(dirs.cache_file())
    return content


def get_default_org(client: TQ42Client) -> Optional[Organization]:
    org_list = list_all_organizations(client=client)
    for org in org_list:
        if org.data.default_org and org.data.default_org is True:
            return org
    return None


def get_default_proj(client: TQ42Client, org_id: str) -> Optional[Project]:
    # get default proj id
    project_list = list_all_projects(client=client, organization_id=org_id)

    for proj in project_list:
        if proj.data.default_project and proj.data.default_project is True:
            return proj

    return None



import os
import unittest
from dataclasses import dataclass
from datetime import datetime
from typing import Any
from unittest import mock

from keyring.errors import InitError, NoKeyringError, PasswordSetError, KeyringLocked

from tq42.client import TQ42Client
from tq42.utils import dirs, file_handling
from tq42.utils.environment import ConfigEnvironment
from tq42.utils.misc import (
    save_token,
    get_token,
)
from tq42.utils.token_manager import TokenManager


class TestUtils(unittest.TestCase):
    def setUp(self):
        self.client = TQ42Client()

    @mock.patch("requests.post")
    def test_renew_expiring_token(self, post_mock):
        @dataclass
        class MockResponse:
            json_data: Any
            status_code: Any

            def json(self):
                return self.json_data

        data = {
            "access_token": "fake_access_token",
            "id_token": "fake_id_token",
            "scope": "fake_scopes",
            "expires_in": 86400,
            "token_type": "Bearer",
        }

        post_mock.return_value = MockResponse(json_data=data, status_code=200)

        token_timestamp = "2022-07-21 13:48:26.580403"
        token_timestamp = datetime.strptime(token_timestamp, "%Y-%m-%d %H:%M:%S.%f")
        # force expire timestamp
        file_handling.write_to_file(
            self.client._timestamp_file_path, str(token_timestamp)
        )

        environment = ConfigEnvironment.from_env()
        token_manager = TokenManager(environment)

        success = token_manager.renew_expiring_token()
        # old token timestamp should renew token
        self.assertTrue(success)

        # renew should return false as timestamp has been updated already
        success = token_manager.renew_expiring_token()
        # new token timestamp should NOT renew token
        self.assertFalse(success)

    def test_save_get_token_with_keyring_enabled(self):
        # keyring is working on Mac Sonoma 14.4 and Windows 11
        token_file_path = os.path.join(dirs.testdata(), "keyring_test.json")
        save_token(
            service_name="tq42_access_token",
            backup_save_path=token_file_path,
            token="test_token",
        )
        token = get_token(
            service_name="tq42_access_token", backup_save_path=token_file_path
        )
        self.assertEqual(token, "test_token")

    @mock.patch("keyring.set_password")
    @mock.patch("keyring.get_password")
    def test_save_get_token_has_InitError(self, mock_set_password, mock_get_password):
        # keyring not working on Ubuntu 20.04.6LTS and causes InitError exception
        token_file_path = os.path.join(dirs.testdata(), "keyring_test.json")
        mock_set_password.side_effect = InitError()
        mock_get_password.side_effect = InitError()
        save_token(
            service_name="tq42_access_token",
            backup_save_path=token_file_path,
            token="test_token",
        )
        token = get_token(
            service_name="tq42_access_token", backup_save_path=token_file_path
        )
        os.remove(token_file_path)
        self.assertEqual(token, "test_token")

    @mock.patch("keyring.set_password")
    @mock.patch("keyring.get_password")
    def test_save_get_token_has_NoKeyringError(
        self, mock_set_password, mock_get_password
    ):
        # keyring not working on Debian and Redhat, NoKeyringError is raised
        token_file_path = os.path.join(dirs.testdata(), "keyring_test.json")
        mock_set_password.side_effect = NoKeyringError()
        mock_get_password.side_effect = NoKeyringError()
        save_token(
            service_name="tq42_access_token",
            backup_save_path=token_file_path,
            token="test_token",
        )
        token = get_token(
            service_name="tq42_access_token", backup_save_path=token_file_path
        )
        os.remove(token_file_path)
        self.assertEqual(token, "test_token")

    @mock.patch("keyring.get_password")
    @mock.patch("keyring.set_password")
    def test_should_return_token_even_if_keyring_locked_and_passwort_set_error_thrown(
        self, mock_set_password, mock_get_password
    ):
        token_file_path = os.path.join(dirs.testdata(), "keyring_test.json")
        mock_set_password.side_effect = PasswordSetError()
        mock_get_password.side_effect = KeyringLocked()

        save_token(
            service_name="tq42_access_token",
            backup_save_path=token_file_path,
            token="test_token",
        )

        token = get_token(
            service_name="tq42_access_token", backup_save_path=token_file_path
        )
        os.remove(token_file_path)
        self.assertEqual(token, "test_token")



import time
import unittest

from tq42.experiment_run import ExperimentRun, HardwareProto
from tq42.functional_tests.functional_test_config import FunctionalCLITestConfig


class TestFunctionalInvokeCliExperimentRunGroup(
    unittest.TestCase, FunctionalCLITestConfig
):
    def test_exp_run_list_should_error_if_no_exp_run_id_is_given(self):
        result = self.runner.invoke(self.cli_entry, ["exp", "run", "list"])
        assert result.exit_code == 2

    def test_exp_run_list(self):
        result = self.runner.invoke(
            self.cli_entry, ["exp", "run", "list", "--exp=" + self.exp]
        )
        assert result.exit_code == 0
        assert result.output is not None

    def test_exp_run_create(self):
        parameters = '{"parameters": {"n": 2, "r": 1, "msg": "correct"}, "inputs": {}}'
        result = self.runner.invoke(
            self.cli_entry,
            [
                "exp",
                "run",
                "create",
                "--exp=" + self.exp,
                "--compute=SMALL",
                "--algorithm=TOY",
                "--version=0.1.0",
                "--parameters=" + parameters,
            ],
        )
        assert result.exit_code == 0
        assert result.output is not None

    def test_exp_run_cancel_should_error_too_late(self):
        result = self.runner.invoke(
            self.cli_entry, ["exp", "run", "cancel", self.exp_run]
        )
        assert result.exit_code == 1

    def test_exp_run_cancel_happy_path(self):
        parameters = {"parameters": {"n": 2, "r": 1, "msg": "correct"}, "inputs": {}}
        exp_run = ExperimentRun.create(
            client=self.client,
            algorithm="TOY",
            version="0.1.0",
            experiment_id=self.exp,
            compute=HardwareProto.SMALL,
            parameters=parameters,
        )

        # we need to wait for a small bit to make sure the experiment run can be retrieved from the system
        time.sleep(5)

        result = self.runner.invoke(
            self.cli_entry, ["exp", "run", "cancel", exp_run.id]
        )
        assert result.exit_code == 0
        assert result.output is not None

    def test_exp_run_check(self):
        result = self.runner.invoke(
            self.cli_entry, ["exp", "run", "check", self.exp_run]
        )
        assert result.exit_code == 0
        assert result.output is not None



from pytest import mark
from tq42.functional_tests.functional_test_config import FunctionalTestConfig


def pytest_addoption(parser):
    parser.addoption(
        "--poll", action="store_true", default=False, help="run polling tests"
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--poll"):
        return
    skip_poll = mark.skip(reason="need --poll option to run")
    for item in items:
        if "poll" in item.keywords:
            item.add_marker(skip_poll)


FunctionalTestConfig.args = FunctionalTestConfig.prepare_defaults()



import unittest
from tq42.experiment_run import ExperimentRun, HardwareProto
from tq42.functional_tests.functional_test_config import FunctionalTestConfig
from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    ExperimentRunStatusProto,
)
from tq42.utils.decorators import timeout


class TestToyAlgorithm(unittest.TestCase, FunctionalTestConfig):
    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    @timeout(180)
    def test_toy_algorithm_successfully_runs(self):
        parameters = {
            "parameters": {
                "n": 2,
                "r": 1,
                "msg": "Toy algorithm test",
            },
            "inputs": {},
        }

        exp_run = ExperimentRun.create(
            client=self.get_client(),
            algorithm="TOY",
            version="0.1.0",
            experiment_id=self.exp,
            compute=HardwareProto.SMALL,
            parameters=parameters,
        )

        self.assertEqual(self.exp, exp_run.data.experiment_id)
        self.assertEqual(ExperimentRunStatusProto.QUEUED, exp_run.data.status)

        final_status = exp_run.poll().data.status
        self.assertEqual(ExperimentRunStatusProto.COMPLETED, final_status)
        self.assertIsNotNone(exp_run.data.result)

    @timeout(180)
    def test_toy_algorithm_fails_with_wrong_message(self):
        parameters = {
            "parameters": {
                "n": 2,
                "r": 1,
                "msg": "wrong",
            },
            "inputs": {},
        }
        exp_run = ExperimentRun.create(
            client=self.get_client(),
            algorithm="TOY",
            version="0.1.0",
            experiment_id=self.exp,
            compute=HardwareProto.SMALL,
            parameters=parameters,
        )

        self.assertEqual(self.exp, exp_run.data.experiment_id)
        self.assertEqual(ExperimentRunStatusProto.QUEUED, exp_run.data.status)

        final_status = exp_run.poll().data.status
        self.assertEqual(ExperimentRunStatusProto.FAILED, final_status)
        self.assertIsNotNone(exp_run.data.error_message)
        self.assertEqual(exp_run.data.error_message, '{"error": "That\'s wrong!"}')


if __name__ == "__main__":
    unittest.main()



import os
import os.path

import tq42.utils.dirs as dirs


def get_current_value(key: str) -> str:
    """
    Get the current value for the provided key in cache file i.e. org, proj..
    If the key was not found raise a KeyError
    """
    if os.path.isfile(dirs.cache_file()):
        with open(dirs.cache_file(), "r") as file:
            for line in file:
                if key in line:
                    split_line = line.strip().split(":")
                    return split_line[1]
    raise KeyError()


def clear_cache() -> None:
    """
    Deletes the content of the cache file.
    """
    # delete context if changing org
    with open(dirs.cache_file(), "w") as file:
        file.truncate()


def write_key_value_to_cache(key: str, value: str) -> None:
    """
    Appends the given tag and id to the cache file
    """
    key_value_line = key + ":" + value

    with open(dirs.cache_file(), "a") as file:
        print(key_value_line, file=file)



from tq42.exceptions.tq42_api_error import TQ42APIError


class AuthenticationError(TQ42APIError):
    """
    Raised when an error occurs during authentication.
    """

    def __init__(self):
        super().__init__("No access token can be retrieved from the response.")



from tq42.exceptions.tq42_api_error import TQ42APIError

_MESSAGE = """
User Not Logged In: Authentication Required
You are not currently logged in. Please log in or authenticate to access this feature or command.
To log in with your username and password or stored token, type the command: tq42 auth login
If you need further assistance, please visit https://help.terraquantum.io/.
""".strip()


class UnauthenticatedError(TQ42APIError):
    """
    Raised when the user is not authenticated
    """

    def __str__(self):
        return _MESSAGE



from tq42.exceptions.tq42_api_error import TQ42APIError


class ExceedRetriesError(TQ42APIError):
    """
    Raised when polling exceeds max retries

    Attributes:
        tries (int): max retries that have been exceeded
    """

    def __init__(self, tries: int):
        self.tries = tries

    def __str__(self):
        return "Polling exceeded. Number of retries: {}".format(self.tries)



import functools
import signal


def timeout(seconds):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            def timeout_handler(signum, frame):
                raise TimeoutError(f"Test timed out after {seconds} seconds")

            signal.signal(signal.SIGALRM, timeout_handler)
            signal.alarm(seconds)
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result

        return wrapper

    return decorator



from tq42.exceptions.tq42_api_error import TQ42APIError


class ExperimentRunCancelError(TQ42APIError):
    """
    Raised when an error occurs while cancelling an experiment run.

    Only raised when the experiment run is either already completed or cancelled.
    """

    def __str__(self):
        return "Cannot cancel a run that was already completed or cancelled."



import os


def read_file(file: str) -> str:
    file_exists = os.path.isfile(file)

    if file_exists is True:
        with open(file, "r") as file:
            token = file.read().rstrip()
            return token

    return ""


def write_to_file(filepath, content):
    content = str(content)
    with open(filepath, "w") as file:
        file.write(content)



import unittest

from google.protobuf.json_format import MessageToDict

from tq42.experiment_run import ExperimentRun, HardwareProto
from tq42.functional_tests.functional_test_config import FunctionalTestConfig
from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    ExperimentRunStatusProto,
)
from tq42.channel import Channel, Ask, Tell
from tq42.utils.decorators import timeout
import OptimizationTestFunctions as otf
import numpy as np
import asyncio


class TestCvaOptAlgorithm(unittest.TestCase, FunctionalTestConfig):
    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    @timeout(240)
    def test_cva_opt_algorithm_successfully_runs(self):
        cva_params = {}
        cva_params["objectives"] = [{"name": "Rosenbrock", "aim_type": "MINIMIZE"}]
        cva_params["variables"] = []
        cva_params["variables"].append(
            {"name": "x1", "info_real": {"lower_bound": -1.0, "upper_bound": 1.0}}
        )
        cva_params["variables"].append(
            {"name": "x2", "info_real": {"lower_bound": -1.0, "upper_bound": 1.0}}
        )
        cva_params["parameters"] = {}
        cva_params["parameters"]["max_generation"] = 50
        cva_params["parameters"]["mue"] = 2
        cva_params["parameters"]["lambda"] = 10

        async def run_exp_with_channel(client, experiment_id, cva_params):
            channel = await Channel.create(client=client)
            # extend cva_params with func_eval_worker_channel_id
            cva_params["func_eval_worker_channel_id"] = channel.id

            # create the experiment run
            run = ExperimentRun.create(
                client=client,
                algorithm="CVA_OPT",
                version="0.1.0",
                experiment_id=experiment_id,
                compute=HardwareProto.SMALL,
                parameters={"parameters": cva_params, "inputs": {}},
            )

            # define the callback function
            async def callback(ask: Ask) -> Tell:
                dim = len(ask.headers)
                func = otf.Rosenbrock(dim)
                y = []
                for parameter in ask.parameters:
                    y.append(float(func(np.array(parameter.values))))
                # add result to data

                return Tell(parameters=ask.parameters, headers=ask.headers, results=y)

            # define a function to be called after the optimization is finished
            def success():
                pass

            # let the channel wait for connections
            is_finished = False
            max_retries = 10
            retries = 0
            while (not is_finished) and (retries < max_retries):
                await channel.connect(
                    callback=callback,
                    finish_callback=success,
                    max_duration_in_sec=None,
                    message_timeout_in_sec=240,
                )
                retries += 1
                run_state = run.check().data.status
                if ExperimentRunStatusProto.Name(run_state) in [
                    "CANCELLED",
                    "COMPLETED",
                    "FAILED",
                    "CANCEL_PENDING",
                ]:
                    is_finished = True
            # return the run to retrieve the result
            return run

        loop = asyncio.get_event_loop()
        run = loop.run_until_complete(
            run_exp_with_channel(self.get_client(), self.exp, cva_params)
        )
        run_result = run.poll()
        self.assertEqual(ExperimentRunStatusProto.COMPLETED, run_result.data.status)
        self.assertIsNotNone(run_result.data.result)
        outcome = MessageToDict(run_result.data.result.outcome)
        self.assertIsNotNone(eval(outcome["result"])["result"]["x1"])


if __name__ == "__main__":
    unittest.main()



import json
import os
import webbrowser
from datetime import datetime

import grpc
from grpc import aio
import requests

from tq42.utils import file_handling, misc
from tq42.utils.token_manager import TokenManager
from tq42.utils.exception_handling import handle_generic_sdk_errors
import time

from com.terraquantum.experiment.v3alpha1.experiment import (
    experiment_service_pb2_grpc as pb2_exp_grpc,
)
from com.terraquantum.experiment.v3alpha2.experimentrun import (
    experiment_run_service_pb2_grpc as pb2_exp_run_grpc,
)
from com.terraquantum.organization.v1.organization import (
    organization_service_pb2_grpc as pb2_org_grpc,
)
from com.terraquantum.project.v1.project import (
    project_service_pb2_grpc as pb2_proj_grpc,
)
from com.terraquantum.storage.v1alpha1 import (
    storage_service_pb2_grpc as pb2_data_grpc,
)
from com.terraquantum.channel.v1alpha1 import (
    channel_service_pb2_grpc as pb2_channel_grpc,
)
import com.terraquantum.plan.v1.plan.plan_service_pb2_grpc as pb2_plan_grpc

from tq42.utils.environment import ConfigEnvironment, environment_default_set
from tq42.exceptions import AuthenticationError

_service_config = {
    "methodConfig": [
        {
            "name": [
                {
                    "service": "com.terraquantum.channel.v1alpha1.ChannelService",
                    "method": "ConnectChannelCustomer",
                }
            ],
            "retryPolicy": {
                "maxAttempts": 5,
                "initialBackoff": "1s",
                "maxBackoff": "10s",
                "backoffMultiplier": 2,
                "retryableStatusCodes": ["UNAVAILABLE", "INTERNAL", "DATA_LOSS"],
            },
        }
    ]
}


class TQ42Client:
    """
    Create a new instance of the TQ42Client to pass to any resource

    Example:
        >>> from tq42.experiment import list_all
        ...
        ... with TQ42Client() as client:
        ...     print(list_all(client=client, project_id="some-project-id"))
    """

    def __call__(self, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __init__(self):
        self._environment = ConfigEnvironment.from_env()
        self._token_manager = TokenManager(self._environment)

        self.server_port = 443

        # instantiate a channel
        self._api_channel = grpc.secure_channel(
            self._environment.api_host,
            grpc.ssl_channel_credentials(),
            options=[
                ("grpc.max_receive_message_length", 10_000_000),
                ("grpc.enable_retries", 1),
                ("grpc.service_config", json.dumps(_service_config)),
            ],
        )
        self.channels_channel = aio.secure_channel(
            self._environment.channels_host, grpc.ssl_channel_credentials()
        )

        # bind the client and the server
        self.organization_client = pb2_org_grpc.OrganizationServiceStub(
            self._api_channel
        )
        self.project_client = pb2_proj_grpc.ProjectServiceStub(self._api_channel)
        self.experiment_client = pb2_exp_grpc.ExperimentServiceStub(self._api_channel)
        self.storage_client = pb2_data_grpc.StorageServiceStub(self._api_channel)
        self.experiment_run_client = pb2_exp_run_grpc.ExperimentRunServiceStub(
            self._api_channel
        )
        self.plan_client = pb2_plan_grpc.PlanServiceStub(self._api_channel)
        self.channel_client = pb2_channel_grpc.ChannelServiceStub(self.channels_channel)

    @handle_generic_sdk_errors
    def login(self):
        """
        Trigger authentication flow. This opens a new browser window to authenticate the sdk.

        If the environment variables `TQ42_AUTH_CLIENT_ID` and `TQ42_AUTH_CLIENT_SECRET` are set the flow is performed without user interaction.
        """
        credential_flow_client_id = os.getenv("TQ42_AUTH_CLIENT_ID")
        credential_flow_client_secret = os.getenv("TQ42_AUTH_CLIENT_SECRET")

        if credential_flow_client_id and credential_flow_client_secret:
            self._login_without_user_interaction(
                client_id=credential_flow_client_id,
                client_secret=credential_flow_client_secret,
            )
        else:
            self._login_with_user_interaction()

    @handle_generic_sdk_errors
    def _login_without_user_interaction(self, client_id: str, client_secret: str):
        response = requests.post(
            self._environment.auth_url_token,
            data=self._environment.client_credentials_data(
                client_id=client_id,
                client_secret=client_secret,
                audience=self._environment.client_credential_flow_audience,
            ),
            headers=self._environment.headers,
        )

        response_json = response.json()
        access_token = response_json.get("access_token")

        if not access_token:
            raise AuthenticationError()

        self._save_access_token(access_token)

    def _login_with_user_interaction(self):
        """
        This method will open a window in your browser where you must confirm the MFA code, then enter your TQ42
        username and password to authenticate. The authentication validity will keep extending as long as you are
        using it within a 30 day period. To access TQ42 services with Python commands, you need a TQ42 account.
        When running TQ42 Python commands, your environment needs to have access to your TQ42 account credentials.
        """
        # Send the POST request and print the response
        response = requests.post(
            self._environment.auth_url_code,
            data=self._environment.code_data,
            headers=self._environment.headers,
        )

        json_response = response.json()
        print(response.text)
        user_code = json_response["user_code"]
        device_code = json_response["device_code"]
        verification_uri_complete = json_response["verification_uri_complete"]
        interval = json_response["interval"]

        # Print the message to the user
        print(
            f"If a browser does not open, please access this URL: {verification_uri_complete} to login. Also check the code: {user_code}"
        )

        webbrowser.open(verification_uri_complete)

        data_token = self._environment.token_data(device_code)

        while True:
            # Send the POST request to get access token and extract the JSON response
            response_token = requests.post(
                self._environment.auth_url_token,
                data=data_token,
                headers=self._environment.headers,
            )
            response_json = response_token.json()

            refresh_token = response_json.get("refresh_token")
            access_token = response_json.get("access_token")

            # If we received an access token, print it and break out of the loop
            if refresh_token and access_token:
                self._save_access_token(access_token)
                self._save_refresh_token(refresh_token)
                break

            # Otherwise, wait for the specified interval before polling again
            time.sleep(interval)

    def _save_access_token(self, access_token: str):
        save_location = misc.save_token(
            service_name="tq42_access_token",
            backup_save_path=self._token_file_path,
            token=access_token,
        )

        print(
            f"Authentication is successful, access token is saved in: {save_location}."
        )

        env_set = environment_default_set(client=self)
        print(env_set)

    def _save_refresh_token(self, refresh_token: str):
        misc.save_token(
            service_name="tq42_refresh_token",
            backup_save_path=self._refresh_token_file_path,
            token=refresh_token,
        )
        current_datetime = datetime.now()
        file_handling.write_to_file(self._timestamp_file_path, current_datetime)

    @property
    def _token_file_path(self):
        return self._token_manager.token_file_path

    @property
    def _timestamp_file_path(self):
        return self._token_manager.timestamp_file_path

    @property
    def _refresh_token_file_path(self):
        return self._token_manager.refresh_token_file_path

    @property
    def metadata(self):
        """
        :meta private:
        """

        self._token_manager.renew_expiring_token()
        token = misc.get_token(
            service_name="tq42_access_token", backup_save_path=self._token_file_path
        )
        return (("authorization", "Bearer " + token),)



import os

from tq42.utils import file_handling
from tq42.utils.environment import ConfigEnvironment
from tq42.utils.misc import get_token, save_token
from datetime import datetime
import requests

_REFRESH_TOKEN_KEY = "tq42_refresh_token"


class TokenManager:
    def __init__(self, environment: ConfigEnvironment):
        self._config_dir = os.path.expanduser("~/.config/tq42")
        self._environment = environment

        if not os.path.exists(self._config_dir):
            os.makedirs(self._config_dir)

    @property
    def token_file_path(self):
        return os.path.join(self._config_dir, "token")

    @property
    def timestamp_file_path(self):
        return os.path.join(self._config_dir, "timestamp")

    @property
    def refresh_token_file_path(self):
        return os.path.join(self._config_dir, "refresh_token")

    def renew_expiring_token(self):
        token_timestamp = file_handling.read_file(self.timestamp_file_path)
        if token_timestamp == "":
            return False
        token_timestamp = datetime.strptime(token_timestamp, "%Y-%m-%d %H:%M:%S.%f")
        diff = datetime.now() - token_timestamp
        # get new access token if token timestamp more than 23 hours(82800 seconds)
        # refresh token is valid for 30 days
        renew_limit = 82800
        if diff.total_seconds() > renew_limit:
            self.request_new_access_token()
            return True

        return False

    def request_new_access_token(self):
        refresh_token = get_token(
            service_name=_REFRESH_TOKEN_KEY,
            backup_save_path=self.refresh_token_file_path,
        )
        data = self._environment.refresh_token_data(refresh_token)
        response = requests.post(
            self._environment.auth_url_token,
            data=data,
            headers=self._environment.headers,
        )
        json_response = response.json()
        if "access_token" in json_response:
            access_token = json_response["access_token"]
            save_token(
                service_name="tq42_access_token",
                backup_save_path=self.token_file_path,
                token=access_token,
            )
            current_datetime = datetime.now()
            file_handling.write_to_file(self.timestamp_file_path, current_datetime)
